// Product automaton -*- c++ -*-

#ifndef PRODUCT_H_
# define PRODUCT_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include "typedefs.h"

# ifdef HASH_MAP_LOC
#  include HASH_MAP_LOC
# else
#  include <map>
# endif // HASH_MAP_LOC
# include <stack>
#ifdef SLIST_LOC
# include SLIST_LOC
#else
# include <slist>
#endif
# include <cassert>

/** @file Product.h
 * Product of reachability graph and the negation of property being verified
 */

/** Product automaton */
class Product
{
public:
  /** Product automaton state */
  class State
  {
  public:
    /** Default constructor */
    State (): rg (0), prop (0) {}
    /** Constructor
     * @param rg	number of the reachability graph state
     * @param prop	number of the property automaton state
     */
    State (card_t rg, unsigned prop) : rg (rg), prop (prop) {}
    /** Copy constructor */
    State (const class State& old) : rg (old.rg), prop (old.prop) {}
    /** Assignment operator */
    class State& operator= (const class State& old) {
      rg = old.rg, prop = old.prop; return *this;
    }
    /** Destructor */
    ~State () {}
    /** Equality comparison */
    bool operator== (const class State& other) const {
      return rg == other.rg && prop == other.prop;
    }
    /** Less-than comparison */
    bool operator< (const class State& other) const {
      return rg < other.rg || (rg == other.rg && prop < other.prop);
    }

    /** Get the hash value */
    size_t operator() () const;

    /** Reachability graph state number */
    card_t rg;
    /** Property automaton state number */
    unsigned prop;
  };
  /** Tarjan state */
  class Tarjan
  {
  public:
    /** Default constructor */
    Tarjan () : state (), depth (0), numSucc (0) {}
    /** Constructor
     * @param state	product automaton state
     * @param depth	search depth
     * @param numSucc	number of enabled gates
     */
    Tarjan (const class State& state,
	    unsigned depth,
	    unsigned numSucc) :
      state (state), depth (depth), numSucc (numSucc) {}
    /** Copy constructor */
    Tarjan (const class Tarjan& old) :
      state (old.state), depth (old.depth), numSucc (old.numSucc) {}
    /** Assignment operator */
    class Tarjan& operator= (const class Tarjan& old) {
      state = old.state, depth = old.depth, numSucc = old.numSucc;
      return *this;
    }
    /** Destructor */
    ~Tarjan () {}

    /** product automaton state */
    class State state;
    /** search depth */
    unsigned depth;
    /** number of enabled gates */
    unsigned numSucc;
  };

  /** Stack of product automaton states */
  typedef std::slist<class State> StateStack;
  /** Stack of states of the strongly connected component algorithm */
  typedef std::slist<class Tarjan> TarjanStack;
# ifdef HASH_MAP_LOC
  struct hasher
  {
    /** Calculate the hash value of a product automaton state
     * @param state	state to be hashed
     * @return		the hash value
     */
    size_t operator() (const class State& state) const {
      return state ();
    }
  };
  /** Map of product automaton states to nonzero search depths */
  typedef std::hash_map<class State,unsigned,struct hasher> StateMap;
# else // HASH_MAP_LOC
  /** Map of product automaton states to nonzero search depths */
  typedef std::map<class State,unsigned> StateMap;
# endif // HASH_MAP_LOC

public:
  /** Constructor
   * @param graph	the reachability graph
   * @param property	the property automaton
   */
  Product (class Graph& graph,
	   const class Property& prop) :
    myGraph (graph), myProp (prop) {}
private:
  /** Copy constructor */
  Product (const class Product& old);
  /** Assignment operator */
  class Product& operator= (const class Product& old);
public:
  /** Destructor */
  ~Product () {}

  /** Check whether the property expressed by the property automaton holds,
   * calculating the product of the reachability graph and the automaton,
   * performing on-the-fly analysis
   * @param state	number of the state where to start the analysis
   * @return		the counterexample path, or NULL if the propery holds
   */
  card_t* analyze (card_t state) const;

# ifdef EXPR_COMPILE
  /** Check whether the property expressed by the property automaton holds,
   * calculating the product of the reachability graph and the automaton,
   * performing on-the-fly analysis
   * @param compilation	the compiled expressions
   * @param state	number of the state where to start the analysis
   * @return		the counterexample path, or NULL if the property holds
   */
  card_t* analyze (const class Compilation& compilation,
		   card_t state) const;
# endif // EXPR_COMPILE

private:
  /** The reachability graph */
  class Graph& myGraph;
  /** The property automaton */
  const class Property& myProp;
};

#endif // PRODUCT_H_
