// Property automaton -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "Property.h"

// #define DEBUG_AUTOMATON // to display the automaton on stdout

#ifdef __WIN32
# undef __STRICT_ANSI__
# include <windows.h>
# include <process.h>
# include <fcntl.h>
# undef pipe
# define pipe(fds) _pipe (fds, 0, _O_TEXT | _O_NOINHERIT)
#endif // __WIN32

#include <cctype>
#ifdef __alpha
# include </usr/include/stdio.h>
# include </usr/include/signal.h>
#elif defined __SUNPRO_CC
using namespace std;
# include <cstdio>
# include <sys/types.h>
# include <signal.h>
#else // not __alpha or __SUNPRO_CC
# include <cstdio>
# include <csignal>
#endif // not __alpha or __SUNPRO_CC
#include <sys/types.h>
#include <unistd.h>

#ifdef __unix
#include <sys/wait.h>
/** Handle a child termination
 * @param pid		process number of the terminated child
 */
extern "C" void childterm (pid_t pid);
/** Handle a signal
 * @param num		number of the signal
 */
extern "C" void sig (int num);
#endif // __unix

#include "PropertyState.h"
#include "Constant.h"
#include "BooleanBinop.h"
#include "NotExpression.h"
#include "LeafValue.h"
#include "BoolType.h"
#include "Net.h"

/** @file Property.C
 * Automaton representing the negation of a property being verified
 */

Property::Property (const class Printer& printer) :
  myPrinter (printer), myFD (0),
  myNumExprs (0), myExprs (0), myNumStates (0),
  myStates (0), myStatesAccept (0),
  myInitialState (UINT_MAX), myNumSets (0)
{
}

Property::~Property ()
{
  delete[] myExprs;
  delete[] myStates;
  delete myStatesAccept;
}

/** Parse a gate condition for a transition
 * @param file		the input stream
 * @param numProps	number of atomic propositions
 * @param props		atomic propositions
 */
static class Expression*
parseGate (FILE* file, unsigned numProps, class Expression** props)
{
  int ch;
#ifdef DEBUG_AUTOMATON
  do
    ch = fgetc (file), fputc (ch, stdout);
  while (isspace (ch));
#else // DEBUG_AUTOMATON
  while (isspace (ch = fgetc (file)));
#endif // DEBUG_AUTOMATON
  switch (ch) {
  case 't':
  case 'f':
    return new class Constant (*new class LeafValue (Net::getBoolType (),
						     ch == 't'));
  case '!':
    if (class Expression* expr = parseGate (file, numProps, props))
      return NotExpression::construct (*expr);
    return 0;
  case '|':
  case '&':
  case 'i':
  case 'e':
    if (class Expression* l = parseGate (file, numProps, props)) {
      if (class Expression* r = parseGate (file, numProps, props)) {
	switch (ch) {
	case '|':
	  return BooleanBinop::construct (false, *l, *r);
	case '&':
	  return BooleanBinop::construct (true, *l, *r);
	case 'i':
	  return BooleanBinop::construct (false,
					 *NotExpression::construct (*l), *r);
	case 'e':
	  return BooleanBinop::construct
	    (false,
	     *BooleanBinop::construct (true, *l, *r),
	     *NotExpression::construct
	     (*BooleanBinop::construct (false, *l->copy (), *r->copy ())));
	}
      }
      l->destroy ();
    }
    return 0;
  case 'p':
    {
      unsigned num;
      if (1 != fscanf (file, "%u", &num))
	fputs ("error in proposition number\n", stderr);
      else if (num > numProps)
	fprintf (stderr, "unknown proposition p%u\n", num);
      else
	return
#ifdef DEBUG_AUTOMATON
	  fprintf (stdout, "%u", num),
#endif // DEBUG_AUTOMATON
	  props[num]->copy ();
    }
    return 0;
  case EOF:
    fputs ("unexpected end of file while parsing formula\n", stderr);
    return 0;
  default:
    fprintf (stderr, "unknown character 0x%02x", ch);
    return 0;
  }
}

bool
Property::create (const char* translator,
		  class Expression& expr)
{
  assert (translator && !myFD && !myNumExprs && !myNumStates);
  /** pipes for the formula and for the automaton */
  int pipe_formula[2], pipe_automaton[2];
  if (pipe (pipe_formula)) {
    perror ("pipe");
    return false;
  }
  if (pipe (pipe_automaton)) {
    perror ("pipe");
    close (pipe_formula[0]), close (pipe_formula[1]);
    return false;
  }
#if defined __WIN32
  /*
   * Since Windows does fork(2)/exec(2) in single pass, we have to
   * duplicate the file descriptors before executing the child process.
   */
  int old_in = _dup (STDIN_FILENO), old_out = _dup (STDOUT_FILENO);
  dup2 (pipe_formula[0], STDIN_FILENO);
  dup2 (pipe_automaton[1], STDOUT_FILENO);

  /* Create the child process */
  HANDLE pid = reinterpret_cast<HANDLE>
    (_spawnlp (_P_NOWAIT, translator, translator, 0));

  _dup2 (old_in, STDIN_FILENO), _dup2 (old_out, STDOUT_FILENO);
  close (old_in), close (old_out);

  if (!pid || pid == INVALID_HANDLE_VALUE) {
    perror (translator);
    close (pipe_formula[0]), close (pipe_formula[1]);
    close (pipe_automaton[0]), close (pipe_automaton[1]);
    return false;
  }
#elif defined __unix
  signal (SIGCHLD, SIG_DFL);
  pid_t pid = fork ();
  if (!pid) {
    /* child process */
    dup2 (pipe_formula[0], STDIN_FILENO);
    dup2 (pipe_automaton[1], STDOUT_FILENO);
    close (pipe_formula[0]), close (pipe_formula[1]);
    close (pipe_automaton[0]), close (pipe_automaton[1]);
    setsid ();
    execlp (translator, translator, 0);
    perror (translator);
    exit (0);
  }
  else if (pid < 0) {
    perror ("fork");
    close (pipe_automaton[0]), close (pipe_automaton[1]);
    close (pipe_formula[0]), close (pipe_formula[1]);
    return false;
  }
#else
# error "unsupported operating system"
#endif

  myFD = pipe_formula[1];
  close (pipe_formula[0]), close (pipe_automaton[1]);
  write (myFD, "!", 1);
  expr.toFormula (*this);
  close (myFD);
  myFD = 0;

  /** file handle for parsing the automaton */
  FILE* f = fdopen (pipe_automaton[0], "r");
  if (!f) {
    perror ("fdopen");
    close (pipe_automaton[0]);
    goto semanticError;
  }

  /** character read from the file */
  int ch;

#ifdef __unix
  for (;;) {
    int status = 0;
    pid_t child = wait (&status);
    if (child == pid) {
      if (!status)
	break;
      fprintf (stderr, "%s returned %d\n", translator, status);
      goto statusError;
    }
    else if (child < 0) {
      perror ("wait");
    statusError:
      signal (SIGCHLD, sig);
      goto semanticError;
    }
    else
      childterm (pid);
  }

  signal (SIGCHLD, sig);
#endif // __unix

  if (2 != fscanf (f, "%u%u", &myNumStates, &myNumSets)) {
  parseError:
    fputs ("error in translated automaton\n", stderr);
  semanticError:
#if defined __WIN32
    TerminateProcess (pid, 0);
#elif defined __unix
    kill (pid, SIGKILL);
#endif // __unix
    delete[] myExprs;
    myNumExprs = 0, myExprs = 0;
    delete[] myStates;
    delete myStatesAccept;
    myNumStates = 0, myStates = 0, myStatesAccept = 0;
    myInitialState = UINT_MAX, myNumSets = 0;
    if (f) fclose (f);
    return false;
  }

  if (!myNumStates) {
    fputs ("the property automaton is empty: tautology?\n", stderr);
    goto semanticError;
  }

  if (myNumSets && myNumStates >= UINT_MAX / myNumSets) {
    fprintf (stderr, "too many states and acceptance sets: %u,%u\n",
	     myNumStates, myNumSets);
    goto semanticError;
  }

  unsigned i;

  if (!(myStates = new class PropertyState[myNumStates])) {
    fprintf (stderr, "cannot allocate %u property automaton states\n",
	     myNumStates);
    goto semanticError;
  }
  if (myNumSets)
    myStatesAccept = new class BitVector (myNumStates * myNumSets);
  else {
    myStatesAccept = new class BitVector (myNumStates);
    for (i = myNumStates; i--; myStatesAccept->assign (i, true));
  }
  if (!myStatesAccept) {
    fprintf (stderr, "cannot allocate %u*%u acceptance sets\n",
	     myNumStates, myNumSets);
    goto semanticError;
  }

  typedef std::map<unsigned,unsigned> NumberMap;
  /** map from state numbers to a continuous sequence */
  NumberMap stateMap;
  /** map from acceptance set numbers to a continuous sequence */
  NumberMap setMap;

#ifdef DEBUG_AUTOMATON
  fprintf (stdout, "%u %u\n", myNumStates, myNumSets);
#endif // DEBUG_AUTOMATON

  for (i = myNumStates; i--; ) {
    unsigned num, initial;
    // state number and "initial state" flag
    if (2 != fscanf (f, "%u%u", &num, &initial) || initial > 1)
      goto parseError;
    std::pair<NumberMap::iterator, bool> p =
      stateMap.insert (NumberMap::value_type (num, stateMap.size ()));

    // translated number of the state
    const unsigned state = p.first->second;

    if (myStates[state].getNumSuccessors ()) {
      fprintf (stderr, "redefinition of state %u\n", num);
      goto semanticError;
    }
    if (initial) {
      if (myInitialState != UINT_MAX) {
	fprintf (stderr, "redefinition of initial state as %u\n", num);
	goto semanticError;
      }
      myInitialState = state;
    }

#ifdef DEBUG_AUTOMATON
    fprintf (stdout, "%u %u ", state, initial);
#endif // DEBUG_AUTOMATON

    // acceptance sets that the state belongs to
    for (;;) {
      while (isspace (ch = fgetc (f)));
      if (ch == '-') {
	if (1 != fscanf (f, "%u", &num) || 1 != num) {
	  fputs ("unexpected data after '-'\n", stderr);
	  goto semanticError;
	}
#ifdef DEBUG_AUTOMATON
	fputs ("-1\n", stdout);
#endif // DEBUG_AUTOMATON
	break;
      }
      ungetc (ch, f);
      if (!isdigit (ch) || 1 != fscanf (f, "%u", &num))
	goto parseError;

      p = setMap.insert (NumberMap::value_type (num, setMap.size ()));
      if (p.second && setMap.size () > myNumSets) {
	fprintf (stderr, "too many acceptance sets: set number %u\n", num);
	goto semanticError;
      }

#ifdef DEBUG_AUTOMATON
      fprintf (stdout, "%u ", p.first->second);
#endif // DEBUG_AUTOMATON
      myStatesAccept->assign (state * myNumSets + p.first->second, true);
    }

    // transitions
    for (;;) {
      while (isspace (ch = fgetc (f)));
      if (ch == '-') {
	if (1 != fscanf (f, "%u", &num) || 1 != num) {
	  fputs ("unexpected data after '-'\n", stderr);
	  goto semanticError;
	}
#ifdef DEBUG_AUTOMATON
	fputs ("-1\n", stdout);
#endif // DEBUG_AUTOMATON
	break;
      }
      ungetc (ch, f);
      if (!isdigit (ch) || 1 != fscanf (f, "%u", &num))
	goto parseError;

      p = stateMap.insert (NumberMap::value_type (num, stateMap.size ()));
      if (p.second && stateMap.size () > myNumStates) {
	fputs ("too many states\n", stderr);
	goto semanticError;
      }

#ifdef DEBUG_AUTOMATON
      fprintf (stdout, "%u", p.first->second);
#endif // DEBUG_AUTOMATON
      if (class Expression* expr = ::parseGate (f, myNumExprs, myExprs))
	myStates[state].addSuccessor (p.first->second, *expr);
      else
	goto parseError;
#ifdef DEBUG_AUTOMATON
      fputc ('\n', stdout);
#endif // DEBUG_AUTOMATON
    }
  }

  while (isspace (ch = fgetc (f)));

  if (ch != EOF) {
    fputs ("extraneous non-whitespace data at end of input\n", stderr);
    goto semanticError;
  }

  if (!myNumSets)
    myNumSets = 1;
  fclose (f);
  return true;
}

void
Property::addOp (enum Property::Op op)
{
  assert (myFD);
  switch (op) {
  case opFalse:
    write (myFD, " f", 2);
    break;
  case opTrue:
    write (myFD, " t", 2);
    break;
  case opNot:
    write (myFD, " !", 2);
    break;
  case opAnd:
    write (myFD, " &", 2);
    break;
  case opOr:
    write (myFD, " |", 2);
    break;
  case opFinally:
    write (myFD, " F", 2);
    break;
  case opGlobally:
    write (myFD, " G", 2);
    break;
  case opNext:
    write (myFD, " X", 2);
    break;
  case opUntil:
    write (myFD, " U", 2);
    break;
  case opRelease:
    write (myFD, " V", 2);
    break;
  }
}

void
Property::addExpression (class Expression& expr)
{
  assert (myFD);
  unsigned i;
  for (i = 0; i < myNumExprs; i++)
    if (myExprs[i] == &expr)
      break;
  if (i == myNumExprs) {
    class Expression** exprs = new class Expression*[i + 1];
    assert (myExprs ? i : !i);
    if (myExprs) {
      memcpy (exprs, myExprs, i * sizeof *exprs);
      delete[] myExprs;
    }
    (myExprs = exprs)[myNumExprs++] = &expr;
  }
  char num[23];
  write (myFD, num, snprintf (num, sizeof num, " p%u", i));
}
